<?php
/*
 * © Copyright 2006, 2007, 2008, 2009 IntraHealth International, Inc.
 * 
 * This File is part of I2CE
 * 
 * I2CE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * @package I2CE
 * @author Luke Duncan <lduncan@intrahealth.org>
 * @since v1.0.0
 * @version v2.0.0
 */
/**
 * Abstract class for objects using the {@link I2CE_Template} class that interact with a database and HTML form.
 * 
 * This class has a few default functions that are used when interacting with the template engine
 * in setting up form values and displaying the data for objects that tie directly to database tables.
 *
 * @package I2CE 
 * @access public
 * @see I2CE_Template
 */
class I2CE_Form extends I2CE_Fuzzy implements Iterator {

    /**
     * @var string The name of this form for the database.
     */
    private $form_name;

    /**
     * @var string The text name of this form for display.
     */
    private $display_name;

    /**
     * @var array The list of fields with all the information about each field.
     */
    protected $fields;

    /**
     * @var mixed String or boolean (false) The form and id  of the parent record for this form if there is one.  Of the form "$form|$id"
     */
    protected $parent;

    /**
     * @var mixed The record id for this entry.
     */
    protected $id;
    
    /**
     * An array of children objects for this form.
     * 
     * It is an array of arrays.  The first being an associative array with the name of the form, the second
     * level is a simple array of the objects.
     * @var array
     */
    public $children;

    /**
     * @var array A list of attributes for this form.
     */
    protected $attributes;

    /**
     * Return the {@link I2CE_FormField form field} value from the {@link $fields} array.
     * @param string $key
     * @return mixed
     */ 
    protected function __get( $key ) {
        if ( array_key_exists( $key, $this->fields ) ) {
            return $this->fields[$key]->getValue();
        } else {
            return null;
        }
    }

    /**
     * Set the {@link I2CE_FormField form field} value in the {@link $fields} array.
     * @param string $key
     * @param mixed $value
     */
    protected function __set( $key, $value ) {
        if ( array_key_exists( $key, $this->fields ) ) {
            $this->fields[$key]->setValue( $value );
        }
    }

    /**
     * Check to see if a {@link I2CE_FormField form field} value has been set.
     * @param string $key
     * @return boolean
     */
    protected function __isset( $key ) {
        if ( array_key_exists( $key, $this->fields ) ) {
            return $this->fields[$key]->issetValue();
        } else {
            return false;
        }
    }

    /**
     * Unset a {@link I2CE_FormField form field}.
     * @param string $key
     */
    protected function __unset( $key ) {
        if ( array_key_exists( $key, $this->fields ) ) {
            $this->fields[$key]->unsetValue();
        }
    }

    /**
     * Rewind the internal pointer for the {@link $fields} array for iterating through the form fields.
     */
    public function rewind() {
        reset( $this->fields );
    }
    /**
     * Return the current value for the {@link $fields} array for iterating through the form fields.
     * @return mixed
     */
    public function current() {
        return current( $this->fields );
    }
    /**
     * Return the current key for the {@link $fields} array for iterating through the form fields.
     * @return mixed
     */
    public function key() {
        return key( $this->fields );
    }
    /**
     * Advance the internal pointer for the {@link $fields} array for iterating through the form fields.
     */
    public function next() {
        next( $this->fields );
    }
    /**
     * Check to see if the current internal pointer for the {@link $fields} array is valid.
     * @return boolean
     */
    public function valid() {
        return array_key_exists( key( $this->fields ), $this->fields );
    }

    /**
     * Get the registered child forms for this form.
     * @returns array The list of child form names registered for this form.
     */
    public function getChildForms() {
        return self::getChildFormsByForm( $this->form_name );
    }

    /**
     * Get the registered child forms for the given form.
     * @param string $form The form to get the children of.
     * @return array The list of child form names registered for the form.
     */
    public static function getChildFormsByForm( $form ) {
        $childForms = array();
        I2CE::getConfig()->setIfIsSet($childForms,"/modules/forms/forms/$form/meta/child_forms",true);
        return $childForms;
    }
        
    /**
     * Get the added children for this form
     * @param string $form   Defaults to null, in which case we get all children.  
     *                       otherwise we get the popoluated children with the specified form
     * @returns array   If $form is null, it is an array with keys the form names and values array 
     *                 of the children of that form type.  If $form is
     *                 set, then it is an array of the forms of type $form
     */
    public function getChildren( $form = null ) {
        if ( is_string($form) && strlen($form) > 0 ) {
            if (array_key_exists($form,$this->children) && is_array($this->children[$form])) {
                return $this->children[$form];
            } else {
                return array();
            }
        } else {
            return $this->children;
        }
    }


    /**
     * @var protected array parent_forms. The array with keys form names and values which are arrays of
     * form names, the form names which the form's parent id can take values in
     */
    protected static $parent_forms  = array();




    /**
     * Gets the allowed parent forms for a given form
     * @param string $form
     * @param boolean $use_cache.  Defaults to true in which case we cached result of finding parent forms
     * @returns array of string, the form names.
     */
    public static function getAllowedParentForms($form, $use_cache = true) {
        if (!$use_cache || !array_key_exists($form,self::$parent_forms)) {
            $parent_forms = array();
            $formsConfig = I2CE::getConfig()->modules->forms->forms;
            foreach ($formsConfig as $pform=>$formConfig) {
                if (!$formConfig instanceof I2CE_MagicDataNode) {
                    continue;
                }
                $pforms = array();
                $formConfig->setIfIsSet($pforms,"meta/child_forms",true);
                if (!in_array($form,$pforms)) {
                    continue;
                }
                $parent_forms[] = $pform;
            }

            self::$parent_forms[$form] =$parent_forms;
        }
        return self::$parent_forms[$form];
    }


    /**
     * Get the names of all the fields added in this form
     * @returns array of string, the field names.
     */
    public function getFieldNames() {
        return array_keys($this->fields);
    }

    /**
     * Create a new instance of a I2CE_Form object.
     * @param string $form The name of this form
     * @param integer $id
     */
    public function __construct( $form, $id='0' ) {
        I2CE_ModuleFactory::callHooks( 'form_pre_construct', 
                array( 'form' => $this, 'form_name' => $form, 'id' => $id ) );
        $this->fields = array();
        $this->children = array();
        $this->setId($id);
        $this->parent = false;
        $this->attributes = array();
        if ( !$form || $form == "" ) {
            I2CE::raiseError( "Blank form name when creating a new I2CE_Form.", E_USER_ERROR );
        }
        $this->setForm( $form );
        $formConfigBase = I2CE::getConfig()->modules->forms->formClasses;
        $factory = I2CE_FormFactory::instance();
        if ( I2CE_ModuleFactory::instance()->isEnabled( "forms-storage" ) ) {
            $this->setChangeType();
        }
        $data = array();
        $this->getFormFieldsData( get_class( $this ), $data );
        $this->addFormFields( $data );
        I2CE_ModuleFactory::callHooks( 'form_post_construct', 
                array( 'form' => $this, 'form_name' => $form, 'id' => $id ) );
    }
    

    protected function getFormFieldsData($class, &$data) {
        if (!empty($class) &&  $class != 'I2CE_Form') {
            //suppose this form has inheritince: I2CE_FormB extends I2CE_FormA extends I2CE_Form
            //calling recusrively at the beginning means that for a field, $field
            //we start at I2CE_Form and look for field data for $field
            //next we look  at I2CE_FormA and look for field data for $field and overwrite anything from I2CE_Form with merge recrusive
            //next we look  at I2CE_FormB and look for field data for $field and overwrite anything from I2CE_Form and I2CE_FormA with merge recrusive
            $this->getFormFieldsData( get_parent_class($class),$data );
        }
        $fieldConfigs = I2CE::getConfig()->traverse("/modules/forms/formClasses/$class/fields",false);
        if ( !$fieldConfigs instanceof I2CE_MagicDataNode) {            
            return;
        }
        foreach ($fieldConfigs  as $field=>$fieldConfig) {
            if (! $fieldConfig instanceof I2CE_MagicDataNode) {
                I2CE::raiseError("Expected non-scalar magic data node at: " . $fieldConfigs->getPath(false) . "/" . $field . " was not found");
                continue;
            }
            if(array_key_exists($field,$data) && is_array($data[$field])) {
                I2CE_Util::merge_recursive($data[$field], $fieldConfig->getAsArray());                
            }  else {
                $data[$field] = $fieldConfig->getAsArray();
            }
        }
    } 

    protected function addFormFields($data) {
        //first we create all the fields and add them to ourself
        foreach ($data as $field=>$fieldData) {
            $this->addField( $field, $fieldData );
            if (!array_key_exists($field,$this->fields) || !$this->fields[$field] instanceof I2CE_FormField) {
                I2CE::raiseError("Could not add field $field to {$this->form_name}");
                unset($this->fields[$field]);
                continue;
            }
        }
        //now that we have created the fields we can set them up and refer to them
        foreach ($data as $field=>$fieldData) {
            if (!array_key_exists($field,$this->fields) || ! $this->fields[$field] instanceof I2CE_FormField) {
                I2CE::raiseError("Could not get field $field");
                continue;
            }
            if (!is_array($fieldData )) {
                I2CE::raiseError("Invalid data for $field of " . $this->form_name);
                $fieldData = array();
            }
            $this->fields[$field]->setDefaultOptions($fieldData);
        }
    }
        
        
    /**
     * Clean up all the fields for this form.
     * 
     * This will unset all the fields associated with this form.  This will remove
     * all circular references to this form so it can be cleaned up by the garbage collector.
     * This should only be called when the form is no longer needed.  Trying to access it
     * after this may cause unexpected results or errors.
     */
    public function cleanup() {
        I2CE_ModuleFactory::callHooks('form_cleanup',array('form'=>$this));
        foreach( $this->fields as $key => $field ) {
            $field = null;
            unset( $this->fields[$key] );
        }
        unset( $this->fields );
    }

    /**
     * Returns the id for this record.
     * @return string
     */
    public function getId() {
        return $this->id;
    }

    /**
     * Returns the form id for this record.
     * @return string of the form "$form|$id"
     */
    public function getFormId() {
        return $this->form_name . '|' . $this->id;
    }


    /**
     * Set the id for this record.
     * @param mixed $id
     */
    public function setId( $id ) {
        if ($id === null || (is_string($id) && strlen($id) == 0)) {
            $id = '0';
        }
        if ( ( strpos( $id, '|' ) ) === false ) {
            $this->id = $id;
        } else {
            list( $form, $id ) = explode( '|', $id, 2 );
            $this->id = $id;
            $this->form_name = $form;
        }
    }

    /**
     * Removes a field from this form.
     * @param string $name 
     * @param array $args The arguments for this field
     */
    public function removeField($name) {
        if (is_scalar($name) && array_key_exists($name,$this->fields)) {
            unset($this->fields[$name]);
        }
    }
        
    /**
     * Adds a field to this form.
     * @param string $name 
     * @param array $args The arguments for this field
     * @returns mixed I2CE_FormField or false on failure
     */
    public function addField( $name, $args ) { 
        if ( !array_key_exists( 'formfield', $args ) || !is_scalar($args['formfield']) ) {
            I2CE::raiseError("No formfield passed to addField", E_USER_WARNING);
            return false;
        }        
        $field = I2CE_FormField::createField($args['formfield'], $name, $args);
        if (!$field) {
            return false;
        }
        $name = $field->getName();
        if ( !$name || $name == "" ) {
            I2CE::raiseError( "Invalid field name for I2CE_Form::addField.", E_USER_NOTICE );
            return false;
        }
        $this->fields[$name] = $field;
        $this->fields[$name]->setForm( $this );
        return $field;
    }



    /**
     * Return the parent id for this record.
     * 
     * If there isn't a parent record set then return the id for this record.
     * @return mixed.  false if there is no parent id or string  a parent id of the form "$form|$id"
     */
    public function getParent( ) {
        return $this->parent;
    }

    /**
     * Return the parent id for this record.
     * 
     * If there isn't a parent record set then return the id for this record.
     * @return mixed.  false if there is no parent id or string  a parent id 
     */
    public function getParentID( ) {
        if (!is_string($this->parent) || strpos($this->parent,'|') === false) {
            return 0;
        } else {
            list($p_form,$p_id) = explode('|',$this->parent,2);
            if (strlen($p_id) == 0) {
                return 0;
            }
            return $p_id;
        }
    }

    /**
     * Return the name of the parent form for this record.
     * 
     * If there isn't a parent record set then return the id for this record.
     * @return mixed.  false if there is no parent   the parent form
     */
    public function getParentForm( ) {
        if (!is_string($this->parent) || strpos($this->parent,'|') === false) {
            return 0;
        } else {
            list($p_form,$p_id) = explode('|',$this->parent,2);
            return $p_form;
        }
    }


        
    /**
     * Set the parent id for this record
     * @param mixed $parent string or I2CE_Form
     * @param boolean $set_id If the parent will only have one child object then you can set the id by setting this to true. 
     */
    public function setParent( $parent, $set_id = false ) {
        if ($parent instanceof I2CE_Form) {
            $this->parent = $parent->form() . '|' . $parent->getID();
        } else {
            if (!is_string($parent) || strlen($parent) == 0) {
                $this->parent = false;
                return false;
            }
            if ((strpos($parent,'|'))=== false) {
                I2CE::raiseError("Bad parent id");
                $this->parent = false;
                return false;
            }
            $this->parent = $parent;
        }
        if ( !$set_id || $this->id != '0'  || !I2CE_ModuleFactory::instance()->isEnabled('forms-storage')) {
            return true;
        }
        if (is_string($parent)) {
            $parent =explode('|',$parent,2);
            list($parent_form_name, $parent_id)  = $parent;
            $parent_form = I2CE_FormFactory::instance()->createForm($parent_form_name.'|'.$parent_id);
            if (!$parent_form instanceof I2CE_Form) {
                I2CE::raiseError("Bad parent form $parent_form_name");
                return false;
            }
        } else {
            $parent_form = $parent; //this has to be an I2CE_Form
        }
        if ($parent_form->getID() == '0' ) {            
            return false;
        }
        $child_ids = $parent_form->getChildIds($this->form_name);                
        if (!is_array($child_ids)) {
            I2CE::raiseError("invalid child form {$this->form_name} for " . $parent_form->form() . " with id " . $parent_form->getID());
            return false;
        } else if  ( count($child_ids) > 1) {
            I2CE::raiseError("No unique child form {$this->form_name} for " . $parent_form->form() . " with id " . $parent_form->getID());
            return false;
        } else if (count($child_ids ) == 1) {
            reset($child_ids);
            $this->setId(current($child_ids));
        }
        return true;
    }

        
    /**
     * Set the form name for this form object.
     * @param string $name
     */
    final public function setForm( $name ) {
        $this->form_name = $name;
    }

    /**
     * Return the form name for this form.
     * @return string
     */
    public function form() {
        return $this->form_name;
    }

    /**
     * Set the display name for this form object.
     * @param string $display
     */
    final public function setDisplayName( $display ) {
        $this->display_name = $display;
    }

    /**
     * Return the display name for this form object.
     * @return string
     */
    public function getDisplayName() {
        return $this->display_name;
    }


    /**
     * Display the field in the given node.
     * @param DOMNode $node
     * @param I2CE_Template $template
     * @param array $args
     */
    public function displayField($node,$template,$args) {
        $field = $this->getField($args[0]);
        if (  $field instanceof I2CE_FormField) {
            $processor = 'processDOM';
            if ($node->hasAttribute('display') && ($field->_hasMethod('processDOM_' . $node->getAttribute('display')))) {
                $processor='processDOM_' . $node->getAttribute('display');
            }
            $field->$processor($node,$template);            
        } 
    }


    /**
     * Process the DOM for this form.
     * @param DOMNode $node
     * @param I2CE_Template $template
     * @param string $method
     * @param array $args
     */
    public function processDOM(&$node,&$template,$method,$args) {
        if (!is_callable(array($this,$method))) {
            I2CE::raiseError("Method $method not callable in {$this->form()}", E_USER_NOTICE);
            //do nothing
            return;
        }
        if ( $node->hasAttribute("ifset") ) {
            $ifset = $node->getAttribute( "ifset" );
            if ( $ifset[0] == "!" ) {
                $not = true;
                $ifset = substr($ifset, 1);
            } else {
                $not = false;
            }
            if  (($ifset == "true") &&  ($method=='displayField')) {
                //we check a field on the current form.  the field name is the argument
                $if_field = $this->getField($args[0]);
            } else {
                $phpfunc = '[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*';
                if (preg_match("/^($phpfunc)\$/",$ifset)) {
                    //we are looking at field of this form
                    $if_field = $this->getField($ifset);
                } else {
                    //we need to check the validity of a field of a different form
                    $if_field = $template->getField($ifset,$node);
                }
            }
            $is_value = true;
            if ( $node->hasAttribute( "ifvalue" ) ) {
                $ifvalue = $node->getAttribute( "ifvalue" );
                if ( $ifvalue[0] == "!" ) {
                    $val_not = true;
                    $ifvalue = substr($ifvalue, 1);
                } else {
                    $val_not = false;
                }
                $is_value = $if_field instanceof I2CE_FormField && $if_field->getDBValue() == $ifvalue;
                if ( $val_not ) $is_value = !$is_value;
            }
            $is_valid = $if_field instanceof I2CE_FormField && $if_field->isValid();
            if ( !( $is_valid  xor $not) || !$is_value )  {
                $node->parentNode->removeChild( $node );
                return;
            }
            $node->removeAttribute( "ifset" );
        } elseif ( $node->hasAttribute( "ifvalue" ) ) {
            $ifvalue = $node->getAttribute( "ifvalue" );
            $if_field = $this->getField($args[0]);
            if ( $ifvalue[0] == "!" ) {
                $val_not = true;
                $ifvalue = substr($ifvalue, 1);
            } else {
                $val_not = false;
            }
            $is_value = $if_field instanceof I2CE_FormField && $if_field->getDBValue() == $ifvalue;
            if ( $val_not ) $is_value = !$is_value;
            if ( !$is_value ) {
                $node->parentNode->removeChild( $node );
                return;
            }
        }
        $this->$method($node,$template,$args);
    }



    /**
     * Reset this object to its original state.
     *
     */
    public function reset() {
        $this->__construct( $this->form() );
    }
        


     /**
      * Add a child form object to this forms list of children.
      * @param I2CE_Form $child_form The child form
      * @param boolean $replace Overwrite the child object if it already exists.
      */
     public function addChildForm( $child_form,  $replace = false ) {
         if (!$child_form instanceof I2CE_Form) {
             return false;
         }
         $form_name = $child_form->form();
         $id = $child_form->getId();
         if (!array_key_exists($form_name,$this->children)) {
             $this->children[$form_name] = array();
         }
         if (  array_key_exists( $id, $this->children[$form_name]) && $this->children[$form_name][$id] instanceof I2CE_Form && !$replace) {
             return;
         }
         $this->children[$form_name][$id] = $child_form;
         $child_form->setParent($this);
     }

    /**
     * Return the I2CE_FormField for the given field name.
     * @param string $field the field name or a form name:field name
     * @return I2CE_FormField
     */
    public function getField( $field ) {
        if (strpos($field,':') !== false) {
            list($form,$field) = explode(':', $field);
        }
        if ( array_key_exists( $field, $this->fields ) ) {
            return $this->fields[ $field ];
        } elseif ( $field == "id" ) {
            $form_field = new I2CE_FormField_STRING_LINE( "id", array( 'in_db' => false ) );
            $form_field->setValue( $this->getFormId() );
            $form_field->setForm( $this );
            return $form_field;
        } elseif ( $field == "parent" ) {
            $form_field = new I2CE_FormField_STRING_LINE( "parent", array( 'in_db' => false ) );
            $form_field->setValue( $this->getParent() );
            $form_field->setForm( $this );
            return $form_field;
        } else {
            return null;
        }
    }
    

    /**
     * Remove fields from the form.
     * 
     * This is to be used when only certain fields of the form are being dealt with so
     * the others can be removed to save processing.
     * @param array $fields The fields being worked with.
     * @param boolean $keep A flag to determine if the given fields should be removed or all except the given fields.
     */
    public function clearFields( $fields, $keep=true ) {
        if ( $keep ) {
            foreach( $this->fields as $key => $field ) {
                if ( !in_array( $key, $fields ) ) {
                    unset( $this->fields[$key] );
                }
            }
        } else {
            foreach( $fields as $key ) {
                unset( $this->fields[$key] );
            }
        }
    }
    
    /**
     * Load the member variables from an array
     * The array can contain the keys 'id', 'parent', 'fields'.  The later of which
     * is an array indexed by field names and which contains the values of the field
     * 
     * @param array $post The post object is passed as a reference
     */
    public function setFromPost($post, $populate_on_set_id = false) {
        if (!is_array($post)) {
            return;
        }
        if (!array_key_exists('fields',$post) || !is_array($post['fields'])) {
            return;
        }
        if (array_key_exists('id',$post['fields']) && is_string($post['fields']['id'])) {
            if (strpos($post['fields']['id'], '|') === false) {
                I2CE::raiseError("Deprecated use of id");
                $id = $post['fields']['id'];
            } else {
                list($form,$id) = explode('|', $post['fields']['id'],2);
                if ($form !== $this->form_name) {
                    I2CE::raiseError("Form name mismatch:" . $form . " != " . $this->form_name);
                    return false;
                }
            }
            if (strlen($id) == 0) {
                $id = '0';
            }
            if ($id !== '0') {
                $this->setID($id);
                if ($populate_on_set_id) {
                    $this->populate();
                }
            }
        }
        if (array_key_exists('parent',$post['fields'])) {
            $this->setParent($post['fields']['parent']);
        }
        foreach ($this->fields as $field=>$fieldObj) {
            if (!array_key_exists($field,$post['fields'])) {
                continue;
            }
            $fieldObj->setFromPost($post['fields'][$field]);
        }
    }

    /**
     * Load the member variables from an array
     * 
     * @param array $post The Post vairables.  Usually from an http request.
     * @param boolean $loadID.  Defaults to true.  If true, and there is
     * exactly one of  entry in the array $post['forms'][$this->form_name], we set the values 
     * of the form's field, id and parent from that entry.  If false, we set the
     * values from $post['forms'][$this->form_name][$this->getID()]  if it is present
     */
    public function load($post, $loadID = true, $populate_on_set_id = true) {        
        if (!is_array($post)
            ||!array_key_exists('forms',$post)
            ||!is_array($post['forms'])
            ||!array_key_exists($this->form_name,$post['forms'])
            ||!is_array($post['forms'][$this->form_name])
            || count($post['forms'][$this->form_name]) == 0){ 
            return;
        }
        if ($loadID) {
            if (count($post['forms'][$this->form_name]) > 1) {
                I2CE::raiseError("Ambiguous form reference when setting from post");
                return;
            } 
            reset($post['forms'][$this->form_name]);
            $this->setFromPost(current($post['forms'][$this->form_name]), true);
        } else {
            if (!array_key_exists($this->getID(),$post['forms'][$this->form_name])
                ||!is_array($post['forms'][$this->form_name][$this->getID()])) {
                return;
            }            
            $this->setFromPost($post['forms'][$this->form_name][$this->getID()],false);
        }
    }


    /**
     * Return all the fields as a query string to be sent to a URL
     * @return string
     */
    public function getQueryString() {
        $query_str = "";
        foreach( $this->fields as $key => $field ) {
            if ( $this->$key != "" )
                $query_str .= "&" . $field->getHTMLName() . "=" . urlencode( $field->getDBValue() );
        }
        return $query_str;
    }

    /**
     * Return the value for the name attribute to be used for this field in a form.
     * @return string
     */
    public function getHTMLName() {
        return $this->form_name;
    }
    /**
     * Return all the fields as an associative array
     * @return array
     */
    public function getQueryFields() {
        $query = array();
        foreach( $this->fields as $key => $field ) {
            if ( $this->$key != "" )
                $query[$field->getHTMLName()] = $field->getDBValue();
        }
        return $query;
    }
    /**
     * Return the values of all the fields that are set.
     * @return string
     */
    public function getQueryDisplay() {
        $display = array();
        foreach( $this->fields as $key => $field ) {
            if ( $field->isInDB() && $field->isValid() && $this->$key != "" )
                $display[] = $field->getDisplayValue();
        }
        if ( sizeof( $display ) > 0 ) {
            return implode( ", ", $display );
        } else {
            return "All";
        }       
    }
    


    /**
     * Validate all fields that are marked as required or unique.
     *
     * This will check all the fields in this form and if they're required or unique it
     * will perform the required checks 
     */
    public function validate() {
        foreach( $this->fields as $field_name => $field_obj ) {
            I2CE_ModuleFactory::callHooks( "validate_formfield",  $field_obj );
        }
    }
        
        
    /**
     *Checks to see if any fields of form has in invalid message
     * @returns boolean
     */
    public function hasInvalid() {
        $invalid = false;
        foreach ($this->fields as $field_obj) {
            $invalid |= $field_obj->hasInvalid();
        }
        return $invalid;
    }




    /**
     * Lookup a given id in the given array.
     * @param integer $id The id to lookup.
     * @param array $arr The array to search through.
     * @return string
     */
    static protected function lookupArray( $id, $arr ) {
        if (!is_array($arr)) {
            I2CE::raiseError("Expected array but not received");
            return "";
        }
        if ( array_key_exists( $id, $arr ) ) {
            return $arr[ $id ];
        } else {
            return "";
        }
    }

        
    /**
     * Set an attribute for this form.
     * @param string $key
     * @param mixed $value
     */
    public function setAttribute( $key, $value ) {
        $this->attributes[$key] = $value;
    }

    /**
     * Return the attribute value for a given attribute.
     * @param string $key
     * @return mixed
     */
    public function getAttribute( $key ) {
        if ( array_key_exists( $key, $this->attributes ) ) {
            return $this->attributes[$key];
        } else {
            return null;
        }
    }

    /**
     * Return true if a given attribute exists for this form.
     * @param string $key
     * @return boolean;
     */
    public function hasAttribute( $key ) {
        return array_key_exists( $key, $this->attributes );
    }







    /**
     * @param string $form  The form name.
      *@param mixed $where_data. Either I2CE_MagicDataNode or array. contains the  where clause information about this form or a nested
     * @param array $ordering. An array of fields to order by.  Defaults to the empty array.  Prepend a - to order by in descending order.
     * @param mixed $limit. Defaults to false.  It true, returns only one result.  If an integer it is the numeber of records to limit to.
     *  If it is as an array of two integers, it is the offset and then number of results to limit to.  
     * @returns mixed an array of matching form ids.  However, ff $limit_one is true or 1 or 
     * array ($offset,1) then then we return either the id or false,  if none found or there was an error.
     */
    public  static function search ( $form, $where_data=array(), $ordering = array(), $limit_one = false) {        
        if (is_string($ordering)) {
            $ordering = array($ordering);
        }
        if (!is_array($ordering)) {
            $ordering = array();
        }        
        if (I2CE_ModuleFactory::instance()->isEnabled('forms-storage')) {
            return  I2CE_FormStorage::search($form,$where_data, $ordering, $limit);
        } else {
            $limit_one = false;
            if (is_array($limit)) {
                if (count($limit) == 2) {
                    end($limit);
                    if (current($limit) == 1) {
                        $limit_one = true;
                    }
                }
            } else {
                $limit_one = (($limit === true ) || (is_numeric($limit) && $limit == 1 ));
            }
            if ($limit_one) {
                return false;
            } else {
                return array();
            }
        }
    }


    /**
     * @param string $form The form name
     * @param array $fields of string. The fields we want returned
      *@param mixed $where_data. Either I2CE_MagicDataNode or array. contains the  where clause information about this form or a nested
     * @param array $ordering. An array of fields to order by.  Defaults to the empty array.  Prepend a - to order by in descending order.
     * @param mixed $limit Defaults to false.  It true, returns only one result.  If an integer it is the numeber of records to limit to.
     *  If it is as an array of two integers, it is the offset and then number of results to limit to.
     * @returns mixed an array with key id's and value and array of values.  the array of values has as keys the fields with their corresponding value.
     */
    public  static function listFields($form, $fields, $where_data=array(), $ordering=array(), $limit = false) {                                                      
        if (is_string($ordering)) {
            $ordering = array($ordering);
        }
        if (!is_array($ordering)) {
            $ordering = array();
        }        
        if (I2CE_ModuleFactory::instance()->isEnabled('forms-storage')) {
            return  I2CE_FormStorage::listFields($form,$fields, false, $where_data, $ordering, $limit);
        } else {
            return array();
        }
    }




}

# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
