<?php
/**
 * @copyright © 2007, 2008, 2009 Intrahealth International, Inc.
 * This File is part of I2CE
 *
 * I2CE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * @package I2CE
 * @author Luke Duncan <lduncan@intrahealth.org>
 * @since v1.0.0
 * @version v2.0.0
 */
/**
 * Base object for dealing with lists of data.
 * 
 * This is an abstract object with base code for many of the I2CE_Form interface methods.  It is used
 * for many of the list database objects such as {@link AcademicLevel} and {@link District}.
 * 
 * It assumes all lists have a code field and by default a description field but this can be overwritten
 * by certain objects like {@link Cadre}.
 * 
 * @package I2CE
 * @abstract
 * @access public
 */
abstract class I2CE_List extends I2CE_Form {
        
    /**
     * The main field name used for display a description of a record.
     */
    const MAIN_FIELD = "name";
    /**
     * The secondary field name used for displaying a description of a record in combination with the MAIN_FIELD.
     */
    const SEC_FIELD = null;
    /**
     * The sort field name to be used for sorting the display list.  This can't be used with the SEC_FIELD option for display.
     */
    const SORT_FIELD = null;
        
    /**
     * An array to cache lookup entries
     * 
     * Any lookups done to the database will be cached in this static array so additional
     * lookups using the same id won't have to access the database.
     * @static
     * @var array
     */
    protected static $cache = array();
        


    /**
     * returns the posible componentization of a form|id 
     * @param string $qry the query need to get the form|id
     * @param array $forms an array of string which are componentized
     * @parm string $component The component we possibly wish to componentize at.
     * @returns string. They query need to turn $qry into one componentized to $component.
     */
    public static function componentizeQuery($qry,$forms,$component) {
        if (count($forms) == 0) {
            return $qry;
        }
        //there are parent forms.  we need to examine this vlaue to
        //see if it is a componentized form.
        foreach ($forms as &$form) {
            $form = '\'' . mysql_real_escape_string($form)  . '\'';
        }
        $form_qry = 'SUBSTRING(' . $qry . ",1, LOCATE('|'," . $qry . ") - 1)";        
        $l_qry = 'IF ( ' . $form_qry . ' IN (' . implode(',',$forms) . '), CONCAT( ' . $qry . ",'@" . mysql_real_escape_string($component) . "')," . $qry . ')';
        return "IF ( LOCATE('|'," . $qry . ") > 0 , " .  $l_qry . ',' . $qry .")";
    }

        
    /**
     * Return the HTML file name for the form template for this form.
     * @return string
     */
    public function getHTMLTemplate() {
        return "lists_form_" . $this->form() . ".html";
    }
    /**
     * Return the list edit type for this list.
     * 
     * The possible return values are "list," "dual," or "select." Select will display a drop
     * down of all choices and list and dual will list them all in a table.  Dual includes the
     * linked list object for the object.
     * @return string
     */
    public function getListType() { return "select"; }
        
    /**
     * Sets the field values to be displayed in the HTML template.
     * @param I2CE_Template &$template
     */
    public function setDisplayData( &$template ) {
        parent::setDisplayData( $template );
        $template->setDisplayData( "add_type", array("type"=>$this->form()) );
        $template->setDisplayData( "type_name", $this->getDisplayName() );
    }
        
    /**
     * Return the display name for this list object.
     * 
     * This will return the same value as lookup() but from the current object instead of from the database.
     * @param string $delim the delimiter between fields.
     * @return string
     */
    public function name( $delim = ' - ' ) {
        $factory = I2CE_FormFactory::instance();
        $main_field = $factory->getConst( $this->form(), "MAIN_FIELD" );
        $sec_field = $factory->getConst( $this->form(), "SEC_FIELD" );
        $name = '';
        if (array_key_exists($main_field,$this->fields) && $this->fields[$main_field] instanceof I2CE_FormField) {
            $name = $this->fields[$main_field]->getDisplayValue();
        } else {
            I2CE::raiseError("Invalid main field $main_field");
        }
        if (is_string($sec_field) && array_key_exists($sec_field,$this->fields) && $this->fields[$sec_field] instanceof I2CE_FormField) {
            $name .= " - " . $this->fields[$sec_field]->getDisplayValue();
        }
        return $name;
    }
        
    /**
     * Return the list of options for this list as an array.
     * 
     * @param string $form_name The form being listed.  This needs to be a subclass of I2CE_List
     * @param string $opt_field An optional field to further limit the list of choices 
     * @param integer $opt_value If the $opt_field is used then this is the value to limit it by.
     * @param array $sub_fields of string.  If $op_value is is set, it is an array of  linked ($form+)$field's to 
     * include results under the optional value.  e.g. if $opt_value was 'country|10' and $sub_fields was
     * array(county+district,district+region,[region],country) we would display all counties and district who
     * are under country|10.  In this case we display the extended version of the option
     * @param boolean $show_hidden.  Show the hidden members of the list, defaults to false.
     * @return array
     */
    static public function listOptions ( $form_name, $opt_field = "", $opt_value = '|', $sub_fields =array() , $selectable_forms=array(), $show_hidden ) {            
        if ( !I2CE_ModuleFactory::instance()->isEnabled( "forms-storage" ) ) {
            return array();
        }
        if ($opt_field && strpos($opt_value,'|') === false) {
            $opt_value = '|';
        }
        if (is_string($sub_fields)) {
            $sub_fields = explode(':',$sub_fields);
        }
        if (!is_array($sub_fields)) {
            $sub_fields =array();
        }
        if (!is_array($selectable_forms)) {
            $selectable_forms =array();
        }
        $factory = I2CE_FormFactory::instance();

        if (!$opt_field || $opt_value == '|'  || count($selectable_forms) == 0 ) {
            $sub_fields = array();
        } else {
            //if we are here, we have subfields to deal with.
            //we need to go through all of our subfields to find the where the form of the optional value matches:
            //$sub_fields =  array(county+district,district+region,[region],country) we would display all counties and district who
            //if $opt_value = 'country|10' we ened with sub_fields array(county+district,district+region,[region])
            //  (if there are selectable forms remaining), then we want to monstermash everything in or under country|10
            //if $opt_value = 'region|10' we ened with sub_fields array(county+district,district+region)
            //  (if there are selectable forms remaining), then we want to monstermash everything in or under region|10
            list($opt_form,$opt_id) = explode('|',$opt_value,2);
            $sub_form = false;
            while (count($sub_fields) > 0) {           
                $sub_field = array_pop($sub_fields);
                if (strlen($sub_field) == 0) {
                    continue;
                }
                if ($sub_field[0] == '[' && $sub_field[strlen($sub_field)-1] == ']') {
                    $sub_field = substr($sub_field,1,strlen($sub_field) -2);
                }
                if (strpos($sub_field,'+') !== false) {
                    list($sub_form,$sub_link) = explode('+',$sub_field,2);
                } else {
                    $sub_form = $sub_field;
                    $sub_link = $sub_field;
                }
                $selectable_forms = array_diff($selectable_forms,array($sub_form)); //remove this form from the selectable list
                if ($sub_form == $opt_form) { 
                    break;
                }
            }
            if (count($sub_fields) == 0) {
                $selectable_forms = array();
            }            
        }
        if ( count($selectable_forms) == 0) {        
            $where = array();
            if ( $opt_field && $opt_value != '|'   ) {
                $where = array(
                    'operator'=>'FIELD_LIMIT',
                    'field'=>$opt_field,
                    'style'=>'equals',
                    'data'=>array(
                        'value'=>$opt_value
                        )
                    );
            }
            $where = self::showHiddenLimit($where,$show_hidden);
            $options = I2CE_FormStorage::listDisplayFields( $form_name, I2CE_List::getDisplayFields($form_name), false, $where, I2CE_List::getSortFields($form_name));
            foreach ($options as &$option) {
                $option = implode(' - ', $option);
            }
            return $options;
        } else {            
            //we have selectable forms underneath this selected value we need to display
            return    self::monsterMash( $form_name,   $opt_field, $opt_value, $sub_fields, array(),true,$show_hidden);
        }
    }


    
    /**
     *  Modifies a where clause to limit to hidden fields as neccesary
     * @param array $where
     * @param boolean $show_hideden
     * @returns array()
     */    
    protected static function showHiddenLimit($where,$show_hidden) {
        if (!$show_hidden) {
            $no_hidden = array(
                'operator'=>'OR',
                'operand'=>array(
                    array(
                        'operator'=>'FIELD_LIMIT',
                        'field'=>'i2ce_hidden',
                        'style'=>'no',
                        'data'=>array( )
                        ),
                    array(
                        'operator'=>'FIELD_LIMIT',
                        'field'=>'i2ce_hidden',
                        'style'=>'null',
                        'data'=>array( )
                        )

                    )
                );
            if (count($where) == 0) {
                $where = $no_hidden;
            } else { 
                $where = array(
                    'operator'=>'AND',
                    'operand'=>array($where, $no_hidden)
                    );
            }
        }
        return $where;
    }

    /**
     * Gets the id's for a form that live at
     * @param string $form_name.  Example person
     * @param array $display_fields.  The fields which we use to populate $options with.  Defaults to an empty array in which case the value of $options is the display fields registered for this form .  
     * @param stirng $opt_field.  The field we are limiting $form_name on.  Example 'residence'
     * @param stirng $opt_formid The form|id and we wish to look that $opt_field takes values in
     * @param artary $sub_fields .  Example  array(county+district,district+region,[region]) If a non-empty array it are the sub-fields path we wish to add to the results.
     * In the example, if $opt_form|$opt_id = country|10 and $link_field = country, then we add in person where residence = country|10, then those person where 
     * residence is an region where  region+country = country|10.  
     * Then we add in those person id's where residence is any distrinct whose region is a region with region+country = country|10.
     * Then we add in those person id's where residence is an county whose district has rgion field a region whose country field is country|10.  
     * @param array $display_fields.  The fields which we return the values of.  Defaults to an empty array in which case the value of $options is the 
     * display fields registered for this form .  
     * @param string $use_display value.  Defaults to true in which case the displayed fields are returned based on the display value.  If false, then the displa
     * @ returns array.  .  Array with keys the ids of $form_name. For values, if $use_dislpay_value is true, then
     * @param boolean $show_hidden.  Show the hidden members of the list, defaults to false.
     */
    public static  function monsterMash($form_name,  $opt_field, $opt_formid,$sub_fields,   $display_fields = array(), $use_display_value = true, $show_hidden = false) {
         //  $form_name  //facility
         //           $opt_field, //location
         //             $opt_formid, //country|10
         //             $sub_fields, //array(county+district,district+region,region+country,country)            
         //             $field_name, //name


            //if $opt_value = 'region|10' we ened with sub_fields array(county+district,district+region)
            //  (if there are selectable forms remaining), then we want to monstermash everything in or under region|10
            //   starting with opt_form = reiong opt_id = 10.  get all failicty with location = region|10
            //next we get all districts where region field is region|10.
            //next we get all counties where district field is district X
        if (is_string($display_fields)) {
            $display_fields = array($display_fields);
        }
        if (!is_array($display_fields) || count($display_fields) == 0) {
            $display_fields =  I2CE_List::getDisplayFields($form_name);
        }
        if (is_string($sub_fields)) {
            $sub_fields = explode(':',$sub_fields);
        }
        if (count($sub_fields) == 0) {
            return array();
        }
        $options =array();    
        list($opt_form,$opt_id) = array_pad(explode('|',$opt_formid,2),2,'');
        self::_monsterMash($options, $form_name,   $opt_field, $opt_form,$opt_id, $sub_fields, '', $display_fields,$use_display_value, $show_hidden);
        return $options;
    }
        
    
    /**
     * Gets the id's for a form that live at
     * @param array &$options.   Where the id's are saved.  Array with keys the ids of $form_name. For values, if $use_dislpay_value is true, then
     * the displayed fields for $form_name that corresponds to that value.  If $use_display_fields is false, then we return an array of the display fields
     * @param string $form_name
     * @param stirng $opt_field.  The field we are limiting $form_name on.
     * @param stirng $opt_form The form we wish to look that $opt_field takes values in
     * @param string $opt_id The id wish to look that $opt_field takes values.  e.g. we are looking for all instances of $form_name where $opt_field  is $opt_form|$opt_id
     * @param string $use_display value.  Defaults to true in which case the displayed fields are returned based on the display value.  If false, then the displa
     * @param artary $sub_fields.  Example  array(county+district,district+region,[region]) 
     * @param string $sub_val.  Defaults to ''.  If use_display_value is set, a string we prepend to the displayed fields for the id
     * @param array $display_fields.  The fields which we use to populate $options with.  Defaults to an empty array in which case the value of $options is the display fields registered for this form .  
     * @param string $use_display value.  Defaults to true in which case the displayed fields are returned based on the display value.  If false, then the displa
     * @param boolean $show_hidden.  Show the hidden members of the list, defaults to false.
     */
    protected static function _monsterMash(
        &$options, $form_name,  $opt_field, $opt_form,$opt_id,$sub_fields, $sub_val = '', 
        $display_fields = array(), $use_display_value = true,
        $show_hidden = false
        ) {
        $where = array(
            'operator'=>'FIELD_LIMIT',
            'field'=>$opt_field,
            'style'=>'equals',
            'data'=>array(
                'value'=>$opt_form .'|' . $opt_id
                )
            );
        $where = self::showHiddenLimit($where,$show_hidden);

        if ($use_display_value) {
            $t_options = I2CE_FormStorage::listDisplayFields( $form_name,$display_fields, false, $where, I2CE_List::getSortFields($form_name));
            foreach ($t_options as $id=>$option) {
                //$option array ar the fields we want
                if ($sub_val) {
                    $options[$id] = $sub_val . ': ' . implode(' - ', $option);
                } else {
                    $options[$id] = implode(' - ', $option);
                }
            }
        } else {
            //append new values to existing ones
            $options +=  I2CE_FormStorage::listFields( $form_name,$display_fields, false, $where, I2CE_List::getSortFields($form_name));
        }
        if (count($sub_fields) == 0) {
            return;
        }
        //we fisrt call this function when $opt_form = 'country' $opt_id = '10'
        //$sub_fields =  array(county+district,district+region,[region]) 
        //$sub_fields =  array(county+district,district+region,[region+country]) 
        $sub_field = array_pop($sub_fields);
        if (strlen($sub_field) == 0) {
            return;
        }
        if ($sub_field[0] == '[' && $sub_field[strlen($sub_field)-1] == ']') {
            $sub_field = substr($sub_field,1,strlen($sub_field) -2);
            $display = false;
        } else {
            $display = true;
        }
        if (strpos($sub_field,'+') !== false) {
            list($sub_form,$sub_link_field) = explode('+',$sub_field,2);
        } else {
            $sub_form =$sub_field; 
            $sub_link_field = $opt_form;
        }
        if ($use_display_value && $display) {
            $sub_val .=      ' ' . self::lookup( $opt_id, $opt_form ); //this looks up the name of country|10
        }
        $sub_where = array(
            'operator'=>'FIELD_LIMIT',
            'field'=>$sub_link_field,
            'style'=>'equals',
            'data'=>array(
                'value'=>$opt_form .'|' . $opt_id
                ));

            //if $opt_value = 'region|10' we ened with sub_fields array(county+district,district+region)
            // start: sub_form = distrinct, sub_field = region
            // next:  sub_form = county, sub_field = district
            //  (if there are selectable forms remaining), then we want to monstermash everything in or under region|10
            //   starting with opt_form = reiong opt_id = 10.  get all failicty with location = region|10
            //next we get all districts where region field is region|10.
            //next we get all counties where district field is district X

            //if $opt_value = 'region|10' we ened with sub_fields array(county+district,district)
            // start: sub_form = distrinct, sub_field = opt_form = region
            // next:  sub_form = county, sub_field = district
            //  (if there are selectable forms remaining), then we want to monstermash everything in or under region|10
            //   starting with opt_form = reiong opt_id = 10.  get all failicty with location = region|10
            //next we get all districts where region field is region|10.
            //next we get all counties where district field is district X
        $sub_where = self::showHiddenLimit($sub_where,$show_hidden);
        $subs = I2CE_FormStorage::search( $sub_form, false, $sub_where, I2CE_List::getSortFields($sub_form) ); //find all regions with country 10
        foreach ($subs as $sub_id) {
            //start: sub_form=distrct, sub_id = districts whose region is region|10
            self::_monsterMash($options,$form_name,$opt_field,$sub_form,$sub_id, $sub_fields, $sub_val, $display_fields,$use_display_value, $show_hidden);
        }
    }

    /**
     * Walk up a form/linked  field path until. Return the id of the top form.
     * $param mixed. Either an array of string or a colon-sepearted string.  (imploded) values is a string of the form:
     *    form1(+field1):...:formM(+fieldM):..:fieldN     
     *    there needs to be at least one colon/array needs to be at least two in length
     * @param string $formid  string of the form "$form|$id".  the starting value.  $form should be one of formX above (e.g. the dbvalue of a map field)
     * @returns mixed. False on failure, string of the form "$form|$id" on success
     */
    public static function walkupFieldPath($field_path,$formid) {
        if (is_string($field_path)) {
            $field_path = explode(':',$field_path);
        }
        if (!is_array($field_path) || count($field_path) < 2) {
            return false;
        }
        list($form,$form_id) = array_pad(explode('|',$formid,2),2,'');
        if (strlen($form_id) == 0 || strlen($form_id) == 0) {
            return false;
        }
        $form_path  = array();
        $link_field_path  = array();
        $len = count($field_path);
        for ($i = $len -1; $i >= 0; $i--) {
            $data = explode('+',$field_path[$i],2);
            if (count($data) == 2) {
                list($form,$link_field) = $data;
            } else {
                $link_field = $form;
                $form = $field_path[$i];
            }
            $form_path[$i] = $form;
            $link_field_path[$i] = $link_field;
        }

        //ksort($form_path);
        //ksort($link_field_path);
        //now we need to see if we can find the form.
        if (($form_index = array_search($form,$form_path)) === false) {
            //the resitrcted form is not among the forms in the form path.
            return false;
        }

        //now we need to walk up the form path until we get to the id of the top most form (e.g. country) for the resitrcted field
        //example:  formid = country|10.  stop this is the top
        //example:  formid = region|10
        //          get dbvalue country field in region|10.  this is the top
        //example:  formid = distrct|10
        //          get dbvalue region field in district|10.  suppose it is refgion|10
        //          get dbvalue country field in region|10.  this is the top
        for ($i= $form_index; $i < $len-1; $i++) {
            //if (($form_id = I2CE_FormStorage::search($form_path[$i],false,$where,array(),1)) === false) {
            if (($formid = I2CE_FormStorage::lookupField($form_path[$i],$form_id,$link_field_path[$i])) === false) {
                //could not walk up the path.
                return false;
            }
            list($form,$form_id) = array_pad(explode('|',$formid,2),2,'');
            if ($form != $link_field_path[$i] || strlen($form_id) == 0 || strlen($form_id) == 0) {
                //be extra sutre that no invalid id received
                return false;
            }
        }
        //$formid is now the id of the top form.
        return $formid;
    }

    /**
     * Returns a list of fields used for displaying this list.
     * @param string $form_name
     * @return array
     */
    static public function getDisplayFields( $form_name ) {
        $factory = I2CE_FormFactory::instance();
        $main_field = $factory->getConst( $form_name, "MAIN_FIELD" );
        $sec_field = $factory->getConst( $form_name, "SEC_FIELD" );
        if (!$main_field) {
            return array();
        }
        $fields = array( $main_field );
        if ( is_string( $sec_field ) ) {
            $fields[] = $sec_field;
        }
        return $fields;
    }

    /**
     * Returns a list of fields used for sorting this list.
     * @param string $form_name
     * @return array
     */
    public static function getSortFields($form_name) {
        $sort_fields = I2CE_List::getDisplayFields($form_name);
        $sort_field = I2CE_FormFactory::instance()->getConst( $form_name, "SORT_FIELD" );
        if ( is_string( $sort_field ) ) {
            array_unshift($sort_fields,  $sort_field);
        }
        return $sort_fields;
    }

        
    /**
     * Looks up the description of the item based on the code.
     * 
     * This is the default method that most implementations of {@link lookup()} use.  It finds the description of
     * the object based on the code and saves it in the {@link cache} and returns it.
     * @param integer $id The code of the entry to lookup.
     * @param string $form_name The name of the form in the database.
     * @return string
     */
    static public function lookup( $id, $form_name ) {
        if ( !$id || $id == "" ) return "";
        if ( !self::isCached( $form_name, $id ) ) {
            $value = I2CE_FormStorage::lookupDisplayField( $form_name, $id, self::getDisplayFields( $form_name ) ) ;
            if ( $value ) {
                self::addCache( $form_name, $id, $value );
                return $value;
            }

        }
        return self::getCache( $form_name, $id );
    }
    
    
    
    /**
     * Checks to see if the {@link id} number is cached.
     * 
     * Checks the {@link cache} to see if {@link code} has been cached or not.
     * @param string $table_name The name of the table in the database.
     * @param integer $code The code of the entry to lookup.
     * @return boolean
     */
    static final protected function isCached( $table_name, $id ) {
        if ( array_key_exists( $table_name, self::$cache ) ) {
            return array_key_exists( $id, self::$cache[ $table_name ] );
        } else {
            return false;
        }
    }
    /**
     * Adds the {@link id} number to the {@link cache}.
     * 
     * @param string $table_name The name of the table in the database.
     * @param integer $code The code of the entry to add.
     * @param string $value The value to add to the cache
     */
    static final protected function addCache( $table_name, $id, $value ) {
        if ( !array_key_exists( $table_name, self::$cache ) ) {
            self::$cache[ $table_name ] = array();
        }
        self::$cache[ $table_name ][ $id ] = $value;       
    }
    /** 
     * Return the {@link cache cached} entry for the {@link id} number.
     * 
     * @param string $table_name The name of the table in the database.
     * @param integer $code The code of the entry to lookup.
     * @return string The value of the cached code.
     */
    static final protected function getCache( $table_name, $id ) {
        if ( self::isCached( $table_name, $id ) ) {
            return self::$cache[ $table_name ][ $id ];
        } else {
            return "";
        }
    }

    /**
     * get an exended display for a field value
     * @param array $fields, An ordered array E.g array('village+county','county','district,'region+country','country').
     * it is an "bottom up" array of string where strings are of the form "$form" or "$form+$link_field".  In the case of
     * the former type, then $link_field is assumed to be the next form.  So for example, "county" has link field "district".
     * If a "$form(+$link_field)" is surrounded by brackets [ ] , it is not displayed.
     * @param string $start_value. E.g. "village|10"
     * @param string $major_separator.  The seperator between the start display value and the extended display values.  If null, we return an array.
     * @param string $minor_separator.  The sepeartor between the extended display values
     * @returns string or array
     */
    public static function getExtendedDisplay($fields, $start_value, $major_separator = ' - ', $minor_separator = ' ') {
        //Example0:
        //$fields == array('village','county','[district+cssc_region],'cssc_region+cssc_country','country')
        //Example1:
        //$fields == array('village+county','county','district,'region+country','country')  -- you could just as easily have used 'region'
        //Example2:
        //$fields == array('village','county');
        //$this->name == county
        if (strpos($start_value,'|') === false) {
            return '';
        } 
        list($start_form,$id) = explode('|',$start_value,2);
        if (!$id || !$start_form  ) {
            if ($major_separator === null) {
                return array();
            } else {
                return '';
            }
        }        
        $ret = I2CE_List::lookup($id,$start_form);  //$ret = 'My Village'
        if ($major_separator === null) {
            $ret = array($ret);
        }
        if (!is_array($fields) || count($fields) == 0) {
            return $ret;
        }
        if (count($fields) == 1 ) {
            return $ret;
        }
        reset($fields);
        $field = array_shift($fields);
        if ( ($pos = strpos($field,'+')) !== false) {            
            list($last_form,$last_link_field) = explode('+',$field,2);
        } else {
            $last_form = $field;
            $last_link_field = false;
        }
        //Example0:$last_form = village, $last_link_field = false
        //Example1:$last_form = village, $last_link_field = county
        $appends = array();
        $started = false;
        while (count($fields) > 0) {
            if (!$started) {
                //start_form == 'village'  b/c $start_value == 'village|10'
                $started = ( $last_form == $start_form);
            }
            $field = array_shift($fields);
            if (!is_string($field)) {
                break;
            }
            //start0: $last_form = village $last_link_field = false $field = 'county' $id = 10
            //next0: $last_form = county $last_link_field = false $field = district+cssc_region $id =20
            //next0: $last_form = distcit $last_link_field = cssc_region $field = cssc_region+cssc_country $id = 30
            //end0: $last_form = cssc_region $last_link_field = cssc_country $field = country $id = 400
            //start1: $last_form = village $last_link_field = county $field = 'county' $id = 10
            //next1: $last_form = county $last_link_field = cssc_region  $field = district $id = 20
            //next1: $last_form = district $last_link_field = false $field = region+country $id = 30
            //end1:$last_form = refion $last_link_field = country $field = country $id = 40
            //start2=end2: $last_form = village $last_link_field = false  $field = county $id = 10            
            $len = strlen($field);
            if ($len >= 2 && $field[0] == '[' && $field[$len-1] == ']') {
                $field = substr($field,1,$len-2);
                $displayedFields = false;
            } else {
                $displayedFields = true;
            }
            if ( ($pos = strpos($field,'+')) !== false) {
                list($form,$link_field) = explode('+',$field,2);
            } else {
                $form = $field;
                $link_field = false;
            }
            if (!$last_link_field) {
                $last_link_field = $form;
            }
            if (!$started) {
                //suppose $start_Value == 'county|20'   so $start_form = county
                //start0: $id = 20, $form = village $link_field = false
                //start1: $id = 20, $form = village $link_field = county
                $last_form = $form; 
                $last_link_field = $link_field;
                //start0: $last_form = villate, $last_link_field = false $id = 20
                //start1: $last_form = village, $last_link_field = county $id = 20
                //we end up in the same state
                continue;
            }
            
            //start0: $form = county $link_field = false  $last_form = village $last_link_field = county $id =10
            //next0: $form = district $link_field = cssc_refion $last_form = county $last_link_field = district $id=20
            //next0: $form = cssc_region $link_field = cssc_country $last_form = distcit $last_link_field = cssc_region $id = 30
            //end0: $form = country $link_field = false $last_form = $cssc_region $last_link_field = cssc_country $id = 400
            //start1: $form = county $link_field = false $last_from = village $last_link_field = county $id =10
            //next1: $form = district $link_field = false  $last_form = county $last_link_field = district $id = 20
            //next1: $form = region $link_field = country $last_form = district $last_link_field = region $id = 30
            //end1: $form = country $link_field = false $last_form = $region $last_link_field = country $id = 40
            //start2=end2: $form = county $link_field = false, $last_form = villate $last_link_field = county $id  = 10
            $data = I2CE_FormStorage::lookupField($last_form, $id, $last_link_field );
            $last_form = $form; 
            $last_link_field = $link_field;
            //start0:  $data = county|20
            //next0: $data = district|30
            //next0: $data = cssc_region|400
            //end0: $data = country = 50
            //start1:  $data = county|20
            //next1: $data = district|30
            //next1: $data = region|40
            //end1: $data = country = 50
            //start2=end2: $data = county|20
            if ( ($pos = strpos($data,'|')) === false) {
                break;
            }            
            $read_form = substr($data,0,$pos);
            $id = substr($data,$pos + 1);
            //start0: $read_form = 'county' $id = 20 
            //next0: $read_form = 'district' $id = 30
            //next0: $read_form = 'cssc_region' $id = 400
            //end0: $read_form = 'country' $id = 50
            //start1: $read_form = 'county' $id = 20 
            //next1: $read_form = 'district' $id = 30 
            //next1: $read_form = 'region' $id = 40
            //end1: $read_form = 'country' $id = 50 
            //start2=end2: $read_form = county $id = 20
            if (!$form || !$id || $read_form != $form) { //sanity checks
                break;
            }
            if (!$displayedFields) {
                continue;
            }
            //start0: $form = county  $id = 20  returns
            //next0: $form = district $id = 30
            //next0: $form = cssc_region $id =400
            //end0: $form = country $id = 50
            //start1: $form = county $id =20
            //next1: $form = district $id = 30
            //next1: $form = region $id =40
            //end1: $form = country $id = 50
            //start2=end2: $form = county $id =20
            $appends[]  = I2CE_List::lookup($id,$form);
        }
        if (count($appends) > 0) {
            $appends = array_reverse($appends);
            if ($major_separator === null) {
                $ret = array_merge($ret,$appends);
            } else {
                $ret .= $major_separator . implode($minor_separator, $appends);
            }
        }
        return $ret;
    }



    /**
     * Create a data tree of the selectable forms.  Deisgned to be fed into tree select
     * @param array $fields, an ordered array E.g array('village+county','county','district,'region+country','country').
     * it is an "bottom up" array of string where strings are of the form "$form" or "$form+$link_field".  In the case of
     * the former type, then $link_field is assumed to be the next form.  So for example, "county" has link field "district".
     * If a "$form(+$link_field)" is surrounded by brackets [ ] , it is not displayed.
     * @param array $forms.  An unorderd array of form names whose values we allow to be selected
     * @param array $limits.  An array with keys form names and value limit data
     * @param array $orders. An array with keys form names and values array of field orders for that form.  If the form name has no orders, we use default
     * ordering for that form based on its displayed firelds
     * @param string $major_separator.  The seperator between the start display value and the extended display values.  If null, then the title is an array of values
     * @param string $minor_separator.  The sepeartor between the extended display values
     * @param boolean $show_hidden.  Show the hidden members of the list, defaults to false.
     */
    public static function createDataTree($fields, $forms,$limits,$orders=array(),$display_first = true , $major_separator = ' - ', $minor_separator = ' ',$show_hidden = false )  {
        //$forms = array(village, county) -- order does not matter.
        //$fields -- order matters.
        //Example0:
        //$fields == array('village','county','[district+cssc_region],'cssc_region+cssc_country','country')
        //Example1:
        //$fields == array('village+county','county','district,'region+country','country')  -- you could just as easily have used 'region'
        //Example2:
        //$fields == array('village','county');        

        //build the tree top down.  top  = country in Example 0,1 and top = county in Example 2
        if (!is_array($forms) || !is_array($fields)  || count($fields) == 0) {
            return array();
        }
        $data = array();
        $displayed = array();
        $last_form = false;
        $fields = array_reverse($fields);
        foreach ($fields as &$field) {
            //start0: $form = country $link_field = false   $last_form = false
            //next0: $form = cssc_region $link_field = country $last_form = country 
            //start1: $form = country $link_field = false   $last_form = false
            //next0: $form = region $link_field = country $last_form = country 
            //start2: $form = county $link_field = false   $last_form = false
            //next2: $form = village $link_field = county  $last_form = county  
            if (!is_string($field)) {
                return array();
            }
            $len = strlen($field);
            if ($len >= 2 && $field[0] == '[' && $field[$len-1] == ']') {
                $field = substr($field,1,$len-2);
                $display = false;
            } else {
                $display = true;
            }
            if ( ($pos = strpos($field,'+')) !== false) {
                list($form,$link_field) = explode('+',$field,2);
                if ($last_form == false) { //throw away junk linked field data on the top level form
                    $link_field =false;
                }
            } else {
                $form = $field;
                $link_field = false;
            }
            if (!$form) {
                return array();
            }
            if (!$link_field) {
                $link_field = $last_form;
            }
            $field = array($form,$link_field);
            $displayed[$form] = $display;
            $last_form = $form;
        }
        $phonebook = array(); //indexed by "$form|$id" with values (by reference) the arrays at which contains the 'children' sub-array  for $form|$id node
        $parent_links = array(); //indexed by "$form|$id" with values "$pform|$pid" which is the form/id that "$form|$id" is linked against
        foreach ($fields as $formfield) {
            list ($form, $link_field) = $formfield;
            if (array_key_exists($form,$limits)) {
                $limit = $limits[$form];
            } else {
                $limit = array();
            }
            //if we dont show the hidden list memmber we need to include the limit where i2ce_disabled is false
            $limit = self::showHiddenLimit($limit,$show_hidden);

            $disp_fields = I2CE_List::getDisplayFields($form);
            //start0:  $form = country, $fields = (name)
            //next0:  $form = cssc_region, $fields = (name, cssc_country)
            //next0:  $form =  district $fields = (name, cssc_region)
            //next0: $form = county $field = (name,distrcit)
            //end0: $form = villate $field = (name,county)
            //start1:  $form = country, $fields = (name)
            //next1:  $form = region, $fields = (name, country)
            //next1:  $form =  district $fields = (name, region)
            //etc.
            if (array_key_exists($form,$orders)) {
                $order = $orders[$form];
            } else {
                $order = I2CE_List::getSortFields($form);
            }
            if ($link_field) {
                $field_datas = I2CE_FormStorage::listFields($form,$link_field,false,$limit,$order);            
            } else {
                $field_datas = I2CE_FormStorage::listFields($form,'id',false,$limit,$order);            
            }
            $display_datas = I2CE_FormStorage::listDisplayFields($form,$disp_fields,false,$limit,$order);            
            $link_id = false;
            $last_link = false;
            $selectable = in_array($form,$forms);
            $title = false;
            foreach ($field_datas as $id=>$field_data) {
                $formid =$form . '|'. $id;
                if (!$link_field) { //this should only be the case for the top form
                    $parent = &$data;
                    $title = false;
                } else {
                    //we are not at the top.
                    $link = $field_data[$link_field];
                    unset($field_data[$link_field]);
                    if ($last_link != $link) {
                        if (!array_key_exists($link,$phonebook)) {   
                            //don't know where to put this as a child of the previous one so skip it
                            continue;
                        }
                        $title =false;
                        $last_link = $link;
                        if (!array_key_exists('children',$phonebook[$link])) {
                            $phonebook[$link]['children'] = array();
                        }
                        $parent = &$phonebook[$link]['children'];
                        //now we need to generate the title/extended display data
                        $titles = array();
                        //example: $diplayed == array(country=>true, region=>false,  district=>true, county=>true village => true)
                        //we have $form = district, $formid = district|30, $parent_link= region|40
                        end($displayed);
                        $disp_form = key($displayed);
                        while ( $disp_form !== false && $disp_form !== $form) {
                            prev($displayed);
                            $disp_form = key($displayed);
                        }
                        //we end here either before the beginning of the array or where $disp_form == $form.                    
                        prev($displayed); //we are now at the one before the $form.  if the current form was district, we are now at region
                        $parent_link = $link;
                        while ( ($parent_link) && ($disp_form = key($displayed) !== false)) {
                            //start: $form = distrct $disp_form = region, $display = false  $parent_link = region|40
                            //next: $disp_form = country $display = true $parent
                            $display = current($displayed);                        
                            if ($display && array_key_exists($parent_link,$phonebook)) {
                                $titles[] = $phonebook[$parent_link]['display'];
                            }
                            if (array_key_exists($parent_link,$parent_links)) {
                                $parent_link = $parent_links[$parent_link];
                            } else {
                                $parent_link = false;
                            }
                            prev($displayed);
                        }
                        //we end with $titles = array ("My Country");  or array ("My District" , "My ")
                        if ($major_separator !== null) {                            
                            $title = implode($minor_separator,array_reverse($titles));
                        } else {
                            $title = array_reverse($titles);
                        }
                    } else {
                        if (!array_key_exists($link,$phonebook) || !$phonebook[$link]) {
                            //don't know where to put this as a child of the previous one so skip it
                            continue;
                        }
                    }
                    $parent_links[$formid] = $link;
                }
                //we know where to add this node.  now lets add it
                //$display = implode(' ', $field_data);
                if (!array_key_exists($id,$display_datas)) {
                    continue;
                }
                $display = implode(' ', $display_datas[$id]);
                //make the title
                $child_data =  array('display'=>$display);
                if ($title) {
                    if ($display_first) {
                        if ($major_separator !== null) {
                            $child_data['title']=$display . $major_separator . $title;
                        } else {
                            $child_data['title'] = array($display) + $title;
                                }
                    } else {
                        if ($major_separator !== null) {
                            $child_data['title']=  $title  . $major_separator .$display ;
                        } else {
                            $child_data['title'] = $title + array($display);
                        }
                    }
                }
                if ($selectable) {
                    $child_data['value'] = $formid;
                }
                $parent[] = $child_data;
                end($parent);
                $phonebook[$formid] = &$parent[key($parent)];
            }
        }
        return $data;
    }

    public static function flattenDataTree($data) {
        $list = array();
        self::_flattenDataTree($data,$list);
        return $list;
    }


    protected static function _flattenDataTree($data,& $list) {
        foreach ($data as $d) {
            if (!is_array($d) || !array_key_exists('display', $d)) {
                continue;
            }
            if (array_key_exists('value',$d)) {
                if (array_key_exists('title',$d) ){
                    $display = $d['title'];
                } else {
                    $display = $d['display'];
                }                
                $list[] = array('value'=>$d['value'], 'display'=>$display);
            }
            if (array_key_exists('children',$d) && is_array($d['children'])) {
                self::_flattenDataTree($d['children'],$list);
            }
        }        
    }



}

# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
